//
// menger - visualization of the Menger sponge
// Copyright (C) 1995-2020  Joachim Reichel <joachim.reichel@posteo.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include "glwidget.h"

#include <chrono>
#include <cmath>

#include <GL/glu.h>

#include <QtWidgets/QApplication>
#include <QtGui/QKeyEvent>
#include <QtGui/QMouseEvent>

#include "mainwindow.h"
#include "renderer.h"

const int lowest_level  = 0;
const int highest_level = 6;

GLWidget::GLWidget( MainWindow* mainwindow, QWidget* parent)
  : QOpenGLWidget( parent),
    m_button_state( Qt::NoButton),
    m_mainwindow( mainwindow),
    m_low_level( 3),
    m_level( 1),
    m_screenshot_is_active( false),
    m_screenshot_width( 0),
    m_screenshot_height( 0)
{
    setFocusPolicy( Qt::ClickFocus);
}

void GLWidget::initializeGL()
{
    m_mainwindow->get_renderer()->init();
}

void GLWidget::paintGL()
{
    if( m_screenshot_is_active) {
        Renderer* renderer = m_mainwindow->get_renderer();
        renderer->screenshot(
            m_screenshot_width, m_screenshot_height, width(), height(), m_screenshot_filename);
        m_screenshot_is_active = false;
        renderer->resize( width(), height());
        update();
    } else
        m_mainwindow->get_renderer()->draw();
}

void GLWidget::resizeGL( int w, int h)
{
    m_mainwindow->get_renderer()->resize( w, h);
}

void GLWidget::screenshot( int w, int h, const std::string& filename)
{
    m_screenshot_width = w;
    m_screenshot_height = h;
    m_screenshot_filename = filename;
    m_screenshot_is_active = true;
    update();
}


void GLWidget::mousePressEvent( QMouseEvent* e)
{
    Renderer* renderer = m_mainwindow->get_renderer();
    m_button_state = e->button();
    m_last = e->pos();
    m_last_press = m_last;
    m_level = renderer->get_level();
    renderer->set_level( std::min( m_level, m_low_level));
}

void GLWidget::mouseMoveEvent( QMouseEvent* e)
{
    Renderer* renderer = m_mainwindow->get_renderer();
    QPoint current = e->pos();
    int size = width() < height() ? width()/2 : height()/2;

    if( m_button_state == Qt::LeftButton) {

        double dx  =   current.x() - m_last.x();
        double dy  = -(current.y() - m_last.y());
        double phi = sqrt(dx*dx+dy*dy) / size * M_PI;
        renderer->rotate( dy, -dx, 0, phi);

    } else if( m_button_state == Qt::MidButton) {

        if( abs( current.x() - m_last.x()) > abs( current.y() - m_last.y())) {

            double dx  = current.x() - m_last.x();
            double phi = dx / size * M_PI;
            renderer->rotate( 0, 0, 1, phi);

        } else {

            double p[3], q[3];
            // fake modelview matrix since the rotation is applied *after* the current modelview matrix
            GLdouble modelview[16] = { 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 };
            GLdouble projection[16];
            GLint viewport[4];
            glGetDoublev( GL_PROJECTION_MATRIX, projection);
            glGetIntegerv( GL_VIEWPORT, viewport);
            gluUnProject( m_last_press.x(), height()-m_last_press.y()-1, 0, modelview, projection, viewport, &p[0], &p[1], &p[2]);
            gluUnProject( width()/2,        height()/2                 , 0, modelview, projection, viewport, &q[0], &q[1], &q[2]);
            double s  = exp( 4.0*( current.y() - m_last.y())/size);
            renderer->translate( q[0]-p[0], q[1]-p[1], q[2]-p[2]);
            renderer->scale( s);
            renderer->translate( p[0]-q[0], p[1]-q[1], p[2]-q[2]);
        }

    } else if( m_button_state == Qt::RightButton) {

            double p[3], q[3];
            // fake modelview matrix since the rotation is applied *after* the current modelview matrix
            GLdouble modelview[16] = { 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 };
            GLdouble projection[16];
            GLint viewport[4];
            glGetDoublev( GL_PROJECTION_MATRIX, projection);
            glGetIntegerv( GL_VIEWPORT, viewport);
            gluUnProject( m_last.x(),  height()-m_last.y()-1,  0, modelview, projection, viewport, &p[0], &p[1], &p[2]);
            gluUnProject( current.x(), height()-current.y()-1, 0, modelview, projection, viewport, &q[0], &q[1], &q[2]);
            renderer->translate( q[0]-p[0], q[1]-p[1], q[2]-p[2]);
    }

    std::chrono::system_clock::time_point start = std::chrono::system_clock::now();

    update();

    std::chrono::system_clock::time_point stop  = std::chrono::system_clock::now();
    double duration = std::chrono::duration<double>( stop-start).count();

    if( (duration > 0.1) && (m_low_level > lowest_level)) {

        if( m_low_level <= m_level)
            m_low_level--;
        else
            m_low_level = std::min( m_level-1, 0);
        renderer->set_level( std::min( m_level, m_low_level));

    } else if( (duration < 0.01) && (m_low_level < highest_level)) {

        if( m_low_level < m_level)
            m_low_level++;
        renderer->set_level( std::min( m_level, m_low_level));

    }

    m_mainwindow->message_tmp( renderer->status( width(), height()).c_str());

    m_last = current;
}

void GLWidget::mouseReleaseEvent( QMouseEvent* /*e*/)
{
    Renderer* renderer = m_mainwindow->get_renderer();
    m_button_state = Qt::NoButton;
    renderer->set_level( m_level);
    m_mainwindow->message_tmp( renderer->status( width(), height()).c_str());
    update();
}

void GLWidget::keyReleaseEvent( QKeyEvent* e)
{
    e->ignore();
}
