//
// menger - visualization of the Menger sponge
// Copyright (C) 1995-2020  Joachim Reichel <joachim.reichel@posteo.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include "mainwindow.h"

#include <cassert>

#if defined( HAVE_GLEW)
#  include <GL/glew.h>
#endif // HAVE_GLEW

#include <QtGui/QKeyEvent>
#include <QtWidgets/QApplication>
#include <QtWidgets/QFileDialog>
#include <QtWidgets/QFrame>
#include <QtWidgets/QHBoxLayout>
#include <QtWidgets/QMenuBar>
#include <QtWidgets/QMessageBox>
#include <QtWidgets/QStatusBar>
#include <QtWidgets/QOpenGLWidget>

#include "glwidget.h"
#include "renderer.h"

MainWindow::MainWindow()
  : QMainWindow()
{
    resize( 800, 600);

    // create Renderer
    m_renderer = new Renderer();

    // frame
    QFrame* frame = new QFrame( this);
    setCentralWidget( frame);

    // OpenGL widget
    m_glwidget = new GLWidget( this, frame);

    // horizontal layout
    QHBoxLayout* flayout = new QHBoxLayout( frame);
    flayout->addWidget( m_glwidget);

    setFocusPolicy( Qt::StrongFocus);

    m_menu_level = new QMenu( "Level", this);
    m_menu_mode  = new QMenu( "Mode", this);
    m_menu_misc  = new QMenu( "Misc", this);

    connect( m_menu_level, SIGNAL( aboutToShow()), this, SLOT( update_menus()));
    connect( m_menu_mode,  SIGNAL( aboutToShow()), this, SLOT( update_menus()));
    connect( m_menu_misc,  SIGNAL( aboutToShow()), this, SLOT( update_menus()));

    menuBar()->addMenu( m_menu_level);
    menuBar()->addMenu( m_menu_mode);
    menuBar()->addMenu( m_menu_misc);

    for( int i = 0; i <= max_level; ++i)
        m_action_level[i] = m_menu_level->addAction(
            QString( "Level &%1").arg( i), this, SLOT( level()), QKeySequence( Qt::Key_0 + i));

    m_action_mode_fast = m_menu_mode->addAction(
        "&Fast", this, SLOT( mode()), QKeySequence( Qt::Key_F));
    m_action_mode_exact = m_menu_mode->addAction(
        "&Exact", this, SLOT( mode()), QKeySequence( Qt::Key_E));
    m_action_mode_simple = m_menu_mode->addAction(
        "&Simple", this, SLOT( mode()), QKeySequence( Qt::Key_S));
    m_action_mode_stupid = m_menu_mode->addAction(
        "S&tupid", this, SLOT( mode()), QKeySequence( Qt::Key_T));

    m_action_misc_reset = m_menu_misc->addAction(
        "&Reset", this, SLOT( misc_reset()), QKeySequence( Qt::Key_R));
    m_action_misc_screenshot = m_menu_misc->addAction(
        "Scree&nshot", this, SLOT( misc_screenshot()), QKeySequence( Qt::Key_N));
    m_action_misc_fullscreen = m_menu_misc->addAction(
        "F&ullscreen", this, SLOT( misc_fullscreen()), QKeySequence( Qt::Key_U));
    m_action_misc_use_vbo = m_menu_misc->addAction(
        "Use &VBO", this, SLOT( misc_use_vbo()), QKeySequence( Qt::Key_V));
    m_action_misc_about = m_menu_misc->addAction(
        "&About", this, SLOT( misc_about()), QKeySequence( Qt::Key_A));
    m_menu_misc->addSeparator();
    m_action_misc_exit = m_menu_misc->addAction(
        "E&xit", QApplication::instance(), SLOT( quit()), QKeySequence( Qt::Key_X));

    for( int i = 0; i <= max_level; ++i)
        m_action_level[i]->setCheckable( true);

    m_action_mode_fast  ->setCheckable( true);
    m_action_mode_exact ->setCheckable( true);
    m_action_mode_simple->setCheckable( true);
    m_action_mode_stupid->setCheckable( true);

    m_action_misc_fullscreen->setCheckable( true);
    m_action_misc_use_vbo   ->setCheckable( true);

    for( int i = 0; i <= max_level; i++)
        m_action_level[i]->setData( i);

    m_action_mode_fast  ->setData( MODE_FAST);
    m_action_mode_exact ->setData( MODE_EXACT);
    m_action_mode_simple->setData( MODE_SIMPLE);
    m_action_mode_stupid->setData( MODE_STUPID);

#if !defined( HAVE_PNG) && !defined( HAVE_JPEG)
    m_action_misc_screenshot->setEnabled( false);
#endif // HAVE_PNG || HAVE_JPEG

#ifndef HAVE_GLEW
    m_action_misc_use_vbo->setEnabled( false);
#endif // HAVE_GLEW

    message_tmp( "Ready");
}

void MainWindow::update_menus()
{
    for( int i = 0; i <= max_level; i++)
        m_action_level[i]->setChecked( m_renderer->get_level() == i);

    m_action_mode_fast  ->setChecked( m_renderer->get_mode() == MODE_FAST);
    m_action_mode_exact ->setChecked( m_renderer->get_mode() == MODE_EXACT);
    m_action_mode_simple->setChecked( m_renderer->get_mode() == MODE_SIMPLE);
    m_action_mode_stupid->setChecked( m_renderer->get_mode() == MODE_STUPID);

    m_action_misc_use_vbo->setChecked( m_renderer->get_use_vbo());
    m_action_misc_fullscreen->setChecked( windowState() & Qt::WindowFullScreen);
}

void MainWindow::level()
{
    QAction* action = qobject_cast<QAction*>( sender());
    level( action->data().toInt());
}


void MainWindow::level( int i)
{
    m_renderer->set_level( i);
    m_glwidget->update();
    message_tmp( m_renderer->status( m_glwidget->width(), m_glwidget->height()).c_str());
}


void MainWindow::mode()
{
    QAction* action = qobject_cast<QAction*>( sender());
    mode( action->data().toInt());
}

void MainWindow::mode( int i)
{
    m_renderer->set_mode( static_cast<Mode>( i));
    m_glwidget->update();
    message_tmp( m_renderer->status( m_glwidget->width(), m_glwidget->height()).c_str());
}

void MainWindow::misc_reset()
{
    m_renderer->reset();
    m_glwidget->update();
    message_tmp( m_renderer->status( m_glwidget->width(), m_glwidget->height()).c_str());
}

void MainWindow::misc_screenshot()
{
#if defined( HAVE_PNG) || defined( HAVE_JPEG)
    QString filter = "Images( ";
#if defined( HAVE_PNG)
    filter += "*.png ";
#else // HAVE_PNG
    filter += "*.jpg *.jpeg ";
#endif // HAVE_PNG
    filter.remove( filter.length()-1, 1);
    filter += ")";

    QString selected_filter;
    QString functionname = QFileDialog::getSaveFileName( this, "", ".", filter, &selected_filter);
    if( functionname == "")
        return;

    message( "Computing screenshot ...");
    std::string s( functionname.toLatin1());
    m_glwidget->screenshot( 1600, 1200, s);
    message_tmp( "Computing screenshot done.");
#else // HAVE_PNG || HAVE_JPEG
    assert( false);
#endif // HAVE_PNG || HAVE_JPEG
}

void MainWindow::misc_use_vbo()
{
    m_renderer->set_use_vbo( !m_renderer->get_use_vbo());
    m_glwidget->update();
    message_tmp( m_renderer->status( m_glwidget->width(), m_glwidget->height()).c_str());
}

void MainWindow::misc_fullscreen()
{
    setWindowState( windowState() ^ Qt::WindowFullScreen);
}

void MainWindow::misc_about()
{
    QMessageBox::information( this, "About", QString( "%1 %2\n\n(c) 1995-2020  Joachim Reichel\n    <joachim.reichel@posteo.de>").
        arg( MENGER_PACKAGE).arg( MENGER_VERSION), QMessageBox::Ok);
}

Renderer* MainWindow::get_renderer() const
{
    return m_renderer;
}

GLWidget* MainWindow::get_glwidget() const
{
    return m_glwidget;
}

void MainWindow::keyPressEvent( QKeyEvent* e)
{
    if( e->key() == Qt::Key_Q)
        QApplication::instance()->quit();
    else if( e->key() == Qt::Key_Escape)
        QApplication::instance()->quit();
    else
        QMainWindow::keyPressEvent( e);
}

void MainWindow::keyReleaseEvent( QKeyEvent* e)
{
    QMainWindow::keyPressEvent( e);
}

void MainWindow::message( const QString & message) const
{
    statusBar()->showMessage( message);
}

void MainWindow::message_tmp( const QString & message) const
{
    statusBar()->showMessage( message, 5000);
}

void MainWindow::message_clear() const
{
    statusBar()->clearMessage();
}
