//
// menger - visualization of the Menger sponge
// Copyright (C) 1995-2020  Joachim Reichel <joachim.reichel@posteo.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include <iostream>
#include <cmath>

#if defined( HAVE_GLEW)
#  include <GL/glew.h>
#endif // HAVE_GLEW

#if defined( HAVE_QT)
#  include <QtWidgets/QApplication>
#  include <QtWidgets/QMessageBox>
#  include <QtOpenGL/QGLFormat>
#endif // HAVE_QT

#include "glwidget.h"
#include "mainwindow.h"
#include "renderer.h"

int main( int argc, char* argv[])
{
    int level       = 1;
    Mode mode       = MODE_FAST;
    double theta_x  = 0.0;
    double theta_y  = 0.0;
    double theta_z  = 0.0;
    double dx       = 0.0;
    double dy       = 0.0;
    double dz       = 0.0;
    double s        = 1.0;
    bool fullscreen = false;
    bool screenshot = false;
    int width       = 1600;
    int height      = 1200;
    std::string filename;
#ifdef HAVE_GLEW
    bool use_vbo    = true;
#else // HAVE_GLEW
    bool use_vbo    = false;
#endif // HAVE_GLEW

    for( int i=1; i<argc; i++) {

        if( (strcmp( argv[i], "--help") == 0) ||( strcmp( argv[i], "-h") == 0)) {
            std::cout << "Usage: " << MENGER_PACKAGE << " [-h|--help] [-v|--version]" << std::endl;
            std::cout << "       " << MENGER_PACKAGE << " [-l|--level l] [-m|--mode m] [-t|--transform ...] \\" << std::endl;
            std::cout << "          [-f|--fullscreen] [-s|--screenshot w h filename]" << std::endl;
            std::cout << std::endl;
            std::cout << "  -h --help        display this help message" << std::endl;
            std::cout << "  -v --version     display version number" << std::endl;
            std::cout << std::endl;
            std::cout << "  -l --level l     sets the level to l (0-6)" << std::endl;
            std::cout << "  -m --mode m      sets the mode (0-3)" << std::endl;
            std::cout << "  -t --transform   set transformation values. This options has to be followed by" << std::endl;
            std::cout << "                   exactly seven doubles, namely values for theta_x, theta_y, theta_z," << std::endl;
            std::cout << "                   dx, dy, dz and s." << std::endl;
            std::cout << std::endl;
            std::cout << "  -f --fullscreen  start in fullscreen mode" << std::endl;
            std::cout << std::endl;
            std::cout << "  -s --screenshot w h filename   save screenshot of size w x h as PNG" << std::endl;
#ifdef HAVE_GLEW
            std::cout << "  -a --avoid-vbos  avoid using vertex buffer objects" << std::endl;
#endif // HAVE_GLEW
            std::cout << std::endl;
            return 1;

        } else if( (strcmp( argv[i], "--version") == 0) || (strcmp( argv[i], "-v") == 0)) {
            std::cout << MENGER_PACKAGE << " " << MENGER_VERSION << std::endl;
            return 2;

        } else if( ((strcmp( argv[i], "--level") == 0) || (strcmp( argv[i], "-l") == 0)) && (i<argc-1)) {
            int l = atoi( argv[++i]);
            if( l >= 0 && l <= 6)
                level = l;

        } else if( ((strcmp( argv[i], "--mode") == 0) || (strcmp( argv[i], "-m") == 0)) && (i<argc-1)) {
            Mode m = static_cast<Mode>( atoi( argv[++i]));
            if( m == MODE_STUPID || m == MODE_SIMPLE || m == MODE_EXACT || m == MODE_FAST)
                mode = m;

        } else if( ((strcmp( argv[i], "--transform") == 0) || (strcmp( argv[i], "-t") == 0)) && (i<argc-7)) {
            theta_x   = atof( argv[++i]);
            theta_y   = atof( argv[++i]);
            theta_z   = atof( argv[++i]);
            dx        = atof( argv[++i]);
            dy        = atof( argv[++i]);
            dz        = atof( argv[++i]);
            double s2 = atof( argv[++i]);
            if( s2 > 0)
                s = s2;

        } else if( (strcmp( argv[i], "--fullscreen") == 0) || (strcmp( argv[i], "-f") == 0)) {
            fullscreen = true;

        } else if( ((strcmp( argv[i], "--screenshot") == 0) || (strcmp( argv[i], "-s") == 0)) && (i<argc-3)) {
            screenshot = true;
            int w = atoi( argv[++i]);
            int h = atoi( argv[++i]);
            if( w > 0)
                width = w;
            if( h > 0)
                height = h;
            filename = std::string( argv[++i]);

#ifdef HAVE_GLEW
        } else if( (strcmp( argv[i], "--avoid-vbos") == 0) || (strcmp( argv[i], "-a") == 0)) {
            use_vbo = false;
#endif // HAVE_GLEW

        } else {
            std::cout << "Unknown option: \"" << argv[i] << "\"" << std::endl;
        }
    }

#if defined( HAVE_QT)

    try {

        // Create Qt application
        QApplication* application = new QApplication( argc, argv);

        if( !QGLFormat::hasOpenGL())  {
            QMessageBox::critical( NULL, "Error", "This system has no OpenGL support. Exiting.");
            return 3;
        }

        // Create top level widget
        MainWindow* mainwindow = new MainWindow();
        Renderer* renderer = mainwindow->get_renderer();
        renderer->set_level( level);
        renderer->set_mode( mode);
        renderer->set_use_vbo( use_vbo);

        // Handle options
        if( fullscreen)
            mainwindow->misc_fullscreen();
        renderer->set_orientation_and_position( theta_x, theta_y, theta_z, dx, dy, dz, s);

        if( screenshot) {
#if defined( HAVE_PNG) || defined( HAVE_JPEG)
            GLWidget* glwidget = mainwindow->get_glwidget();
            glwidget->screenshot( width, height, filename);
            mainwindow->showMaximized();
            delete mainwindow;
            delete application;
            return 0;
#else // HAVE_PNG || HAVE_JPEG
            (void) width;  // avoid warning
            (void) height; // avoid warning
            std::cerr << "Screenshots need PNG or JPEG support. Exiting.";
            delete mainwindow;
            delete application;
            return 4;
#endif // HAVE_PNG || HAVE_JPEG
        }

        mainwindow->show();
        int exitcode = application->exec();

        delete mainwindow;
        delete application;

        return exitcode;

    } catch( const std::exception& e) {

       std::cerr << "Exception: " << e.what() << std::endl;
       return 5;

    }

#else // HAVE_QT

    std::cerr << "This program needs Qt." << std::endl;
    return 6;

#endif // HAVE_QT
}
