//
// menger - visualization of the Menger sponge
// Copyright (C) 1995-2020  Joachim Reichel <joachim.reichel@posteo.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include "renderer.h"

#include <cassert>
#include <chrono>
#include <cmath>
#include <sstream>
#include <stdexcept>
#include <iostream>

#if defined( HAVE_GLEW)
#  include <GL/glew.h>
#endif // HAVE_GLEW

#include <GL/glu.h>

#if defined( HAVE_JPEG)
#  include <cstdio>
extern "C" {
#  include <jpeglib.h>
}
#endif // HAVE_JPEG

#if defined( HAVE_PNG)
extern "C" {
#  include <png.h>
}
#endif // HAVE_PNG

#include "tile_rendering.h"

struct Subcube
{
    int x;                       ///< x coordinate
    int y;                       ///< y coordinate
    int z;                       ///< z coordinate
    bool inherited_visibility;   ///< front visibility inherited?
    bool always_visible;         ///< front always visible?
};

Subcube subcubes[20] = {

        // back
        {  0, 0, 0, false, false},
        {  1, 0, 0, false, true },
        {  2, 0, 0, false, false},
        {  0, 1, 0, false, true },
        // 1, 1, 0,
        {  2, 1, 0, false, true,},
        {  0, 2, 0, false, false},
        {  1, 2, 0, false, true,},
        {  2, 2, 0, false, false},

        // middle
        {  0, 0, 1, false, false},
        // 1, 0, 1
        {  2, 0, 1, false, false},
        // 0, 1, 1
        // 1, 1, 1
        // 2, 1, 1
        {  0, 2, 1, false, false},
        // 1, 2, 1
        {  2, 2, 1, false, false},

        // front
        {  0, 0, 2, true , false},
        {  1, 0, 2, true , false},
        {  2, 0, 2, true , false},
        {  0, 1, 2, true , false},
        // 1, 1, 2
        {  2, 1, 2, true , false},
        {  0, 2, 2, true , false},
        {  1, 2, 2, true , false},
        {  2, 2, 2, true , false}
};

const int sections0 = 1;
const int sections1 = 3;
const int sections2 = 9;

const int faces0    = 1;
const int faces1    = 8;
const int faces2f   = 80;
const int faces2e   = 96;

const int face_data0[faces0*12] = {
        0, 0, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1 };

const int face_data1[faces1*12] = {
        1, 0, 1, 2, 0, 1, 2, 1, 1, 1, 1, 1,
        2, 1, 1, 3, 1, 1, 3, 2, 1, 2, 2, 1,
        1, 2, 1, 2, 2, 1, 2, 3, 1, 1, 3, 1,
        0, 1, 1, 1, 1, 1, 1, 2, 1, 0, 2, 1,
        0, 0, 3, 3, 0, 3, 3, 1, 3, 0, 1, 3,
        3, 0, 3, 3, 3, 3, 2, 3, 3, 2, 0, 3,
        3, 3, 3, 0, 3, 3, 0, 2, 3, 3, 2, 3,
        0, 3, 3, 0, 0, 3, 1, 0, 3, 1, 3, 3 };

const int face_data2f[faces2f*12] = {
        0, 0, 1, 9, 0, 1, 9, 1, 1, 0, 1, 1,
        0, 2, 1, 9, 2, 1, 9, 3, 1, 0, 3, 1,
        0, 6, 1, 9, 6, 1, 9, 7, 1, 0, 7, 1,
        0, 8, 1, 9, 8, 1, 9, 9, 1, 0, 9, 1,
        0, 0, 1, 1, 0, 1, 1, 9, 1, 0, 9, 1,
        2, 0, 1, 3, 0, 1, 3, 9, 1, 2, 9, 1,
        6, 0, 1, 7, 0, 1, 7, 9, 1, 6, 9, 1,
        8, 0, 1, 9, 0, 1, 9, 9, 1, 8, 9, 1,
        0, 3, 1, 3, 3, 1, 3, 4, 1, 0, 4, 1,
        0, 5, 1, 3, 5, 1, 3, 6, 1, 0, 6, 1,
        6, 3, 1, 9, 3, 1, 9, 4, 1, 6, 4, 1,
        6, 5, 1, 9, 5, 1, 9, 6, 1, 6, 6, 1,
        3, 0, 1, 4, 0, 1, 4, 3, 1, 3, 3, 1,
        5, 0, 1, 6, 0, 1, 6, 3, 1, 5, 3, 1,
        3, 6, 1, 4, 6, 1, 4, 9, 1, 3, 9, 1,
        5, 6, 1, 6, 6, 1, 6, 9, 1, 5, 9, 1,

        3, 0, 3, 6, 0, 3, 6, 1, 3, 3, 1, 3,
        3, 2, 3, 6, 2, 3, 6, 3, 3, 3, 3, 3,
        3, 0, 3, 4, 0, 3, 4, 3, 3, 3, 3, 3,
        5, 0, 3, 6, 0, 3, 6, 3, 3, 5, 3, 3,
        0, 3, 3, 3, 3, 3, 3, 4, 3, 0, 4, 3,
        0, 5, 3, 3, 5, 3, 3, 6, 3, 0, 6, 3,
        0, 3, 3, 1, 3, 3, 1, 6, 3, 0, 6, 3,
        2, 3, 3, 3, 3, 3, 3, 6, 3, 2, 6, 3,
        6, 3, 3, 9, 3, 3, 9, 4, 3, 6, 4, 3,
        6, 5, 3, 9, 5, 3, 9, 6, 3, 6, 6, 3,
        6, 3, 3, 7, 3, 3, 7, 6, 3, 6, 6, 3,
        8, 3, 3, 9, 3, 3, 9, 6, 3, 8, 6, 3,
        3, 6, 3, 6, 6, 3, 6, 7, 3, 3, 7, 3,
        3, 8, 3, 6, 8, 3, 6, 9, 3, 3, 9, 3,
        3, 6, 3, 4, 6, 3, 4, 9, 3, 3, 9, 3,
        5, 6, 3, 6, 6, 3, 6, 9, 3, 5, 9, 3,

        1, 0, 4, 2, 0, 4, 2, 1, 4, 1, 1, 4,
        1, 2, 4, 2, 2, 4, 2, 3, 4, 1, 3, 4,
        0, 1, 4, 1, 1, 4, 1, 2, 4, 0, 2, 4,
        2, 1, 4, 3, 1, 4, 3, 2, 4, 2, 2, 4,
        7, 0, 4, 8, 0, 4, 8, 1, 4, 7, 1, 4,
        7, 2, 4, 8, 2, 4, 8, 3, 4, 7, 3, 4,
        6, 1, 4, 7, 1, 4, 7, 2, 4, 6, 2, 4,
        8, 1, 4, 9, 1, 4, 9, 2, 4, 8, 2, 4,
        1, 6, 4, 2, 6, 4, 2, 7, 4, 1, 7, 4,
        1, 8, 4, 2, 8, 4, 2, 9, 4, 1, 9, 4,
        0, 7, 4, 1, 7, 4, 1, 8, 4, 0, 8, 4,
        2, 7, 4, 3, 7, 4, 3, 8, 4, 2, 8, 4,
        7, 6, 4, 8, 6, 4, 8, 7, 4, 7, 7, 4,
        7, 8, 4, 8, 8, 4, 8, 9, 4, 7, 9, 4,
        6, 7, 4, 7, 7, 4, 7, 8, 4, 6, 8, 4,
        8, 7, 4, 9, 7, 4, 9, 8, 4, 8, 8, 4,

        0, 0, 7, 9, 0, 7, 9, 1, 7, 0, 1, 7,
        0, 2, 7, 9, 2, 7, 9, 3, 7, 0, 3, 7,
        0, 6, 7, 9, 6, 7, 9, 7, 7, 0, 7, 7,
        0, 8, 7, 9, 8, 7, 9, 9, 7, 0, 9, 7,
        0, 0, 7, 1, 0, 7, 1, 9, 7, 0, 9, 7,
        2, 0, 7, 3, 0, 7, 3, 9, 7, 2, 9, 7,
        6, 0, 7, 7, 0, 7, 7, 9, 7, 6, 9, 7,
        8, 0, 7, 9, 0, 7, 9, 9, 7, 8, 9, 7,
        0, 3, 7, 3, 3, 7, 3, 4, 7, 0, 4, 7,
        0, 5, 7, 3, 5, 7, 3, 6, 7, 0, 6, 7,
        6, 3, 7, 9, 3, 7, 9, 4, 7, 6, 4, 7,
        6, 5, 7, 9, 5, 7, 9, 6, 7, 6, 6, 7,
        3, 0, 7, 4, 0, 7, 4, 3, 7, 3, 3, 7,
        5, 0, 7, 6, 0, 7, 6, 3, 7, 5, 3, 7,
        3, 6, 7, 4, 6, 7, 4, 9, 7, 3, 9, 7,
        5, 6, 7, 6, 6, 7, 6, 9, 7, 5, 9, 7,

        0, 0, 9, 9, 0, 9, 9, 1, 9, 0, 1, 9,
        0, 2, 9, 9, 2, 9, 9, 3, 9, 0, 3, 9,
        0, 6, 9, 9, 6, 9, 9, 7, 9, 0, 7, 9,
        0, 8, 9, 9, 8, 9, 9, 9, 9, 0, 9, 9,
        0, 0, 9, 1, 0, 9, 1, 9, 9, 0, 9, 9,
        2, 0, 9, 3, 0, 9, 3, 9, 9, 2, 9, 9,
        6, 0, 9, 7, 0, 9, 7, 9, 9, 6, 9, 9,
        8, 0, 9, 9, 0, 9, 9, 9, 9, 8, 9, 9,
        0, 3, 9, 3, 3, 9, 3, 4, 9, 0, 4, 9,
        0, 5, 9, 3, 5, 9, 3, 6, 9, 0, 6, 9,
        6, 3, 9, 9, 3, 9, 9, 4, 9, 6, 4, 9,
        6, 5, 9, 9, 5, 9, 9, 6, 9, 6, 6, 9,
        3, 0, 9, 4, 0, 9, 4, 3, 9, 3, 3, 9,
        5, 0, 9, 6, 0, 9, 6, 3, 9, 5, 3, 9,
        3, 6, 9, 4, 6, 9, 4, 9, 9, 3, 9, 9,
        5, 6, 9, 6, 6, 9, 6, 9, 9, 5, 9, 9 };

const int face_data2e[faces2e*12] = {
        1, 0, 1, 2, 0, 1, 2, 1, 1, 1, 1, 1,
        4, 0, 1, 5, 0, 1, 5, 1, 1, 4, 1, 1,
        7, 0, 1, 8, 0, 1, 8, 1, 1, 7, 1, 1,
        0, 1, 1, 1, 1, 1, 1, 2, 1, 0, 2, 1,
        2, 1, 1, 4, 1, 1, 4, 2, 1, 2, 2, 1,
        5, 1, 1, 7, 1, 1, 7, 2, 1, 5, 2, 1,
        8, 1, 1, 9, 1, 1, 9, 2, 1, 8, 2, 1,
        1, 2, 1, 2, 2, 1, 2, 4, 1, 1, 4, 1,
        4, 2, 1, 5, 2, 1, 5, 3, 1, 4, 3, 1,
        7, 2, 1, 8, 2, 1, 8, 4, 1, 7, 4, 1,
        0, 4, 1, 1, 4, 1, 1, 5, 1, 0, 5, 1,
        2, 4, 1, 3, 4, 1, 3, 5, 1, 2, 5, 1,
        6, 4, 1, 7, 4, 1, 7, 5, 1, 6, 5, 1,
        8, 4, 1, 9, 4, 1, 9, 5, 1, 8, 5, 1,
        1, 5, 1, 2, 5, 1, 2, 7, 1, 1, 7, 1,
        4, 6, 1, 5, 6, 1, 5, 7, 1, 4, 7, 1,
        7, 5, 1, 8, 5, 1, 8, 7, 1, 7, 7, 1,
        0, 7, 1, 1, 7, 1, 1, 8, 1, 0, 8, 1,
        2, 7, 1, 4, 7, 1, 4, 8, 1, 2, 8, 1,
        5, 7, 1, 7, 7, 1, 7, 8, 1, 5, 8, 1,
        8, 7, 1, 9, 7, 1, 9, 8, 1, 8, 8, 1,
        1, 8, 1, 2, 8, 1, 2, 9, 1, 1, 9, 1,
        4, 8, 1, 5, 8, 1, 5, 9, 1, 4, 9, 1,
        7, 8, 1, 8, 8, 1, 8, 9, 1, 7, 9, 1,

        3, 0, 3, 6, 0, 3, 6, 1, 3, 3, 1, 3,
        3, 2, 3, 6, 2, 3, 6, 3, 3, 3, 3, 3,
        3, 0, 3, 4, 0, 3, 4, 3, 3, 3, 3, 3,
        5, 0, 3, 6, 0, 3, 6, 3, 3, 5, 3, 3,
        0, 3, 3, 3, 3, 3, 3, 4, 3, 0, 4, 3,
        0, 5, 3, 3, 5, 3, 3, 6, 3, 0, 6, 3,
        0, 3, 3, 1, 3, 3, 1, 6, 3, 0, 6, 3,
        2, 3, 3, 3, 3, 3, 3, 6, 3, 2, 6, 3,
        6, 3, 3, 9, 3, 3, 9, 4, 3, 6, 4, 3,
        6, 5, 3, 9, 5, 3, 9, 6, 3, 6, 6, 3,
        6, 3, 3, 7, 3, 3, 7, 6, 3, 6, 6, 3,
        8, 3, 3, 9, 3, 3, 9, 6, 3, 8, 6, 3,
        3, 6, 3, 6, 6, 3, 6, 7, 3, 3, 7, 3,
        3, 8, 3, 6, 8, 3, 6, 9, 3, 3, 9, 3,
        3, 6, 3, 4, 6, 3, 4, 9, 3, 3, 9, 3,
        5, 6, 3, 6, 6, 3, 6, 9, 3, 5, 9, 3,

        1, 0, 4, 2, 0, 4, 2, 1, 4, 1, 1, 4,
        1, 2, 4, 2, 2, 4, 2, 3, 4, 1, 3, 4,
        0, 1, 4, 1, 1, 4, 1, 2, 4, 0, 2, 4,
        2, 1, 4, 3, 1, 4, 3, 2, 4, 2, 2, 4,
        7, 0, 4, 8, 0, 4, 8, 1, 4, 7, 1, 4,
        7, 2, 4, 8, 2, 4, 8, 3, 4, 7, 3, 4,
        6, 1, 4, 7, 1, 4, 7, 2, 4, 6, 2, 4,
        8, 1, 4, 9, 1, 4, 9, 2, 4, 8, 2, 4,
        1, 6, 4, 2, 6, 4, 2, 7, 4, 1, 7, 4,
        1, 8, 4, 2, 8, 4, 2, 9, 4, 1, 9, 4,
        0, 7, 4, 1, 7, 4, 1, 8, 4, 0, 8, 4,
        2, 7, 4, 3, 7, 4, 3, 8, 4, 2, 8, 4,
        7, 6, 4, 8, 6, 4, 8, 7, 4, 7, 7, 4,
        7, 8, 4, 8, 8, 4, 8, 9, 4, 7, 9, 4,
        6, 7, 4, 7, 7, 4, 7, 8, 4, 6, 8, 4,
        8, 7, 4, 9, 7, 4, 9, 8, 4, 8, 8, 4,

        1, 0, 7, 2, 0, 7, 2, 1, 7, 1, 1, 7,
        4, 0, 7, 5, 0, 7, 5, 1, 7, 4, 1, 7,
        7, 0, 7, 8, 0, 7, 8, 1, 7, 7, 1, 7,
        0, 1, 7, 1, 1, 7, 1, 2, 7, 0, 2, 7,
        2, 1, 7, 4, 1, 7, 4, 2, 7, 2, 2, 7,
        5, 1, 7, 7, 1, 7, 7, 2, 7, 5, 2, 7,
        8, 1, 7, 9, 1, 7, 9, 2, 7, 8, 2, 7,
        1, 2, 7, 2, 2, 7, 2, 4, 7, 1, 4, 7,
        4, 2, 7, 5, 2, 7, 5, 3, 7, 4, 3, 7,
        7, 2, 7, 8, 2, 7, 8, 4, 7, 7, 4, 7,
        0, 4, 7, 1, 4, 7, 1, 5, 7, 0, 5, 7,
        2, 4, 7, 3, 4, 7, 3, 5, 7, 2, 5, 7,
        6, 4, 7, 7, 4, 7, 7, 5, 7, 6, 5, 7,
        8, 4, 7, 9, 4, 7, 9, 5, 7, 8, 5, 7,
        1, 5, 7, 2, 5, 7, 2, 7, 7, 1, 7, 7,
        4, 6, 7, 5, 6, 7, 5, 7, 7, 4, 7, 7,
        7, 5, 7, 8, 5, 7, 8, 7, 7, 7, 7, 7,
        0, 7, 7, 1, 7, 7, 1, 8, 7, 0, 8, 7,
        2, 7, 7, 4, 7, 7, 4, 8, 7, 2, 8, 7,
        5, 7, 7, 7, 7, 7, 7, 8, 7, 5, 8, 7,
        8, 7, 7, 9, 7, 7, 9, 8, 7, 8, 8, 7,
        1, 8, 7, 2, 8, 7, 2, 9, 7, 1, 9, 7,
        4, 8, 7, 5, 8, 7, 5, 9, 7, 4, 9, 7,
        7, 8, 7, 8, 8, 7, 8, 9, 7, 7, 9, 7,

        0, 0, 9, 9, 0, 9, 9, 1, 9, 0, 1, 9,
        0, 2, 9, 9, 2, 9, 9, 3, 9, 0, 3, 9,
        0, 6, 9, 9, 6, 9, 9, 7, 9, 0, 7, 9,
        0, 8, 9, 9, 8, 9, 9, 9, 9, 0, 9, 9,
        0, 0, 9, 1, 0, 9, 1, 9, 9, 0, 9, 9,
        2, 0, 9, 3, 0, 9, 3, 9, 9, 2, 9, 9,
        6, 0, 9, 7, 0, 9, 7, 9, 9, 6, 9, 9,
        8, 0, 9, 9, 0, 9, 9, 9, 9, 8, 9, 9,
        0, 3, 9, 3, 3, 9, 3, 4, 9, 0, 4, 9,
        0, 5, 9, 3, 5, 9, 3, 6, 9, 0, 6, 9,
        6, 3, 9, 9, 3, 9, 9, 4, 9, 6, 4, 9,
        6, 5, 9, 9, 5, 9, 9, 6, 9, 6, 6, 9,
        3, 0, 9, 4, 0, 9, 4, 3, 9, 3, 3, 9,
        5, 0, 9, 6, 0, 9, 6, 3, 9, 5, 3, 9,
        3, 6, 9, 4, 6, 9, 4, 9, 9, 3, 9, 9,
        5, 6, 9, 6, 6, 9, 6, 9, 9, 5, 9, 9 };

const char* mode_to_str( Mode m)
{
    if( m == MODE_STUPID) return "stupid";
    if( m == MODE_SIMPLE) return "simple";
    if( m == MODE_EXACT)  return "exact";
    if( m == MODE_FAST)   return "fast";
    assert( false);
    return "Unknown";
}

Renderer::Renderer()
{
    reset();
}

void Renderer::init()
{
#ifdef HAVE_GLEW
    glewInit();
#endif // HAVE_GLEW

    glEnable( GL_LIGHT0);
    glEnable( GL_LIGHTING);

    glShadeModel( GL_FLAT);

    glEnable( GL_DEPTH_TEST);
    glEnable( GL_NORMALIZE);
    glEnable( GL_COLOR_MATERIAL);
    glEnable( GL_CULL_FACE);

    GLfloat white[4] = { 1.0, 1.0, 1.0, 1.0 };
    glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, white);
}

void Renderer::draw()
{
    if( m_use_vbo) {

        if( m_update_vbo) {

            m_vbo.clear();
            generate();

#ifdef HAVE_GLEW
            GLuint vbo;
            glGenBuffers( 1, &vbo);
            glBindBuffer( GL_ARRAY_BUFFER, vbo);
            glBufferData( GL_ARRAY_BUFFER, m_vbo.size() * sizeof( float), &m_vbo[0], GL_STREAM_DRAW);
            glVertexPointer( 3, GL_FLOAT, 0, NULL);
            glEnableClientState( GL_VERTEX_ARRAY);
#else // HAVE_GLEW
            assert( false);
#endif // HAVE_GLEW

            m_update_vbo = false;
        }

    } else {

        m_vbo.clear();
        std::vector<float> tmp;
        m_vbo.swap( tmp);

    }

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadMatrixd( m_M);

    m_use_vbo ? glDrawArrays( GL_QUADS, 0, m_vbo.size() / 3) : generate();
    glRotated( 90, 0, 1, 0);
    m_use_vbo ? glDrawArrays( GL_QUADS, 0, m_vbo.size() / 3) : generate();
    glRotated( 90, 0, 1, 0);
    m_use_vbo ? glDrawArrays( GL_QUADS, 0, m_vbo.size() / 3) : generate();
    glRotated( 90, 0, 1, 0);
    m_use_vbo ? glDrawArrays( GL_QUADS, 0, m_vbo.size() / 3) : generate();
    glRotated( 90, 1, 0, 0);
    m_use_vbo ? glDrawArrays( GL_QUADS, 0, m_vbo.size() / 3) : generate();
    glRotated( 180, 1, 0, 0);
    m_use_vbo ? glDrawArrays( GL_QUADS, 0, m_vbo.size() / 3) : generate();
}

#if defined( HAVE_PNG) || defined( HAVE_JPEG)
void Renderer::screenshot(
    int width, int height, int tile_width, int tile_height, const std::string& filename)
{
    GLubyte* image = new GLubyte[width * height * 3];

    std::chrono::system_clock::time_point start = std::chrono::system_clock::now();

    TRcontext* tr = trNew();
    trTileSize( tr, tile_width, tile_height, 8);
    trImageSize( tr, width, height);
    trImageBuffer( tr, GL_RGB, GL_UNSIGNED_BYTE, image);

    double ratio = static_cast<double>( width) / static_cast<double>( height);
    if( ratio>1)
        trOrtho( tr, -ratio, ratio, -1, 1, -100, 100);
    else
        trOrtho( tr, -1, 1, -1/ratio, 1/ratio, -100, 100);

    do {
        trBeginTile( tr);
        draw();
    } while( trEndTile( tr));

    trDelete( tr);

    std::chrono::system_clock::time_point stop  = std::chrono::system_clock::now();
    double duration = std::chrono::duration<double>( stop-start).count();
    std::cerr << duration << " seconds" << std::endl;

#if defined( HAVE_PNG)
    write_png_file( reinterpret_cast<const char*>( image), width, height, filename);
#elif defined( HAVE_JPEG)
    write_jpeg_file( reinterpret_cast<const char*>( image), width, height, filename);
#else // HAVE_JPEG
    assert( false);
#endif // HAVE_JPEG

    delete[] image;
}
#endif // HAVE_PNG || HAVE_JPEG

void Renderer::resize( int w, int h) const
{
    glViewport( 0, 0, static_cast<GLsizei>( w), static_cast<GLsizei>( h));
    glMatrixMode( GL_PROJECTION);
    glLoadIdentity();
    double ratio = static_cast<double>( w) / static_cast<double>( h);
    if( ratio>1)
       glOrtho( -ratio, ratio, -1, 1, -100, 100);
    else
        glOrtho( -1, 1, -1/ratio, 1/ratio, -100, 100);
    glMatrixMode( GL_MODELVIEW);
}

void Renderer::reset()
{
    for( int i = 0; i < 16; i++) // NOLINT modernize-loop-convert
        m_M[i] = 0;
    m_M[0] = m_M[5] = m_M[10] = m_M[15] = 1;
    m_level = 1;
    m_mode = MODE_FAST;
    m_nr_of_quads = 0;
#ifdef HAVE_GLEW
    m_use_vbo = true;
#else
    m_use_vbo = false;
#endif // HAVE_GLEW
    m_update_vbo = true;
}

void Renderer::translate( double dx, double dy, double dz)
{
    m_M[12] += m_M[15]*dx;
    m_M[13] += m_M[15]*dy;
    m_M[14] += m_M[15]*dz;
}

void Renderer::scale( double s)
{
    assert( s > 0);
    m_M[15] /= s;
}

void Renderer::rotate( double x, double y, double z, double phi)
{
    double dx = m_M[12] / m_M[15];
    double dy = m_M[13] / m_M[15];
    double dz = m_M[14] / m_M[15];

    translate( -dx, -dy, -dz);

    double length = sqrt( x*x + y*y + z*z);
    if( length <= 0)
        return;

    x = x / length;
    y = y / length;
    z = z / length;

    double sin_phi = sin( phi);
    double cos_phi = cos( phi);

    // column-major order
    GLdouble R[16], tmp[16];
    R[ 0] = (1 - cos_phi)*x*x + cos_phi;
    R[ 1] = (1 - cos_phi)*y*x - sin_phi*z;
    R[ 2] = (1 - cos_phi)*z*x + sin_phi*y;
    R[ 3] = 0;
    R[ 4] = (1 - cos_phi)*x*y + sin_phi*z;
    R[ 5] = (1 - cos_phi)*y*y + cos_phi;
    R[ 6] = (1 - cos_phi)*z*y - sin_phi*x;
    R[ 7] = 0;
    R[ 8] = (1 - cos_phi)*x*z - sin_phi*y;
    R[ 9] = (1 - cos_phi)*y*z + sin_phi*x;
    R[10] = (1 - cos_phi)*z*z + cos_phi;
    R[11] = 0;
    R[12] = 0;
    R[13] = 0;
    R[14] = 0;
    R[15] = 1;

    for( int i = 0; i < 4; i++)
        for( int j = 0; j < 4; j++) {
             tmp[i+4*j] = 0;
             for( int k = 0; k < 4; k++)
                 tmp[i+4*j] += R[i+4*k] * m_M[k+4*j];
         }

    for( int i = 0; i < 16; i++)
        m_M[i] = tmp[i];

    translate( dx, dy, dz);
}

void Renderer::set_orientation_and_position(
    double theta_x, double theta_y, double theta_z, double dx, double dy, double dz, double s)
{
    assert( s > 0);

    double sx = sin( theta_x);
    double sy = sin( theta_y);
    double sz = sin( theta_z);
    double cx = cos( theta_x);
    double cy = cos( theta_y);
    double cz = cos( theta_z);

    m_M[ 0] = cy*cz;
    m_M[ 1] = cy*sz;
    m_M[ 2] = -sy;
    m_M[ 3] = 0;
    m_M[ 4] = sx*sy*cz - cx*sz;
    m_M[ 5] = sx*sy*sz + cx*cz;
    m_M[ 6] = sx*cy;
    m_M[ 7] = 0;
    m_M[ 8] = cx*sy*cz + sx*sz;
    m_M[ 9] = cx*sy*sz - sx*cz;
    m_M[10] = cx*cy;
    m_M[11] = 0;
    m_M[12] = dx;
    m_M[13] = dy;
    m_M[14] = dz;
    m_M[15] = 1/s;
}

void Renderer::get_orientation_and_position(
    double& theta_x, double& theta_y, double& theta_z, double& dx, double& dy, double& dz, double& s) const
{
    theta_y = asin( -m_M[2]);

    if( (m_M[0] == 0) && (m_M[1] == 0) && (m_M[2] == 1)) { // theta_y == -M_PI_2

        theta_x = 0;
        theta_z = -atan2( -m_M[4], m_M[8]);

    } else if( (m_M[0] == 0) && (m_M[1] == 0) && (m_M[2] == -1)) { // theta_y == M_PI_2)

        theta_x = 0;
        theta_z = -atan2( -m_M[4], m_M[8]);

    } else {

        theta_x = atan2( m_M[6], m_M[10]);
        theta_z = atan2( m_M[1], m_M[0]);
    }

    dx = m_M[12];
    dy = m_M[13];
    dz = m_M[14];
    s  = 1 / m_M[15];
}

void Renderer::set_level( int l)
{
    assert( l >= 0 && l <= 6);
    m_update_vbo |= l != m_level;
    m_level = l;
}

int Renderer::get_level() const
{
    return m_level;
}

void Renderer::set_mode( Mode m)
{
    assert( m == MODE_STUPID || m == MODE_SIMPLE || m == MODE_EXACT || m == MODE_FAST);
    m_update_vbo |= m != m_mode;
    m_mode = m;
}

Mode Renderer::get_mode() const
{
    return m_mode;
}

void Renderer::set_use_vbo( bool v)
{
#ifndef HAVE_GLEW
    assert( !v);
#endif // HAVE_GLEW
    m_update_vbo |= !m_use_vbo && v;
    m_use_vbo = v;
}

bool  Renderer::get_use_vbo() const
{
    return m_use_vbo;
}

std::string Renderer::status( int width, int height) const
{
    double theta_x, theta_y, theta_z, dx, dy, dz, s;
    get_orientation_and_position( theta_x, theta_y, theta_z, dx, dy, dz, s);

    std::ostringstream str;
    str << "level: " << m_level << ", ";
    str << "mode: " << mode_to_str( m_mode) << ", ";
    str << "use VBOs: " << (m_use_vbo ? "yes, " : "no, ");
    str << "quads: " << m_nr_of_quads << ", ";
    str << "transformation: " << theta_x << ", ";
    str << theta_y << ", ";
    str << theta_z << ", ";
    str << dx << ", ";
    str << dy << ", ";
    str << dz << ", ";
    str << s << ", ";
    str << "width: " << width << ", ";
    str << "height: " << height;
    return str.str();
}

void Renderer::generate ()
{
    m_nr_of_quads = 0;

    if( !m_use_vbo)
        glBegin( GL_QUADS);

    double size = 1.0;
    double offset = -size/2;

    if( (m_mode == MODE_STUPID) || (m_mode == MODE_SIMPLE))

        recurse( offset, offset, offset, size, true, 0);

    else

        if( m_level == 0)
            leaf( offset, offset, offset, size, true, sections0, faces0, face_data0);

        else if( m_level == 1)
            leaf( offset, offset, offset, size, true, sections1, faces1, face_data1);

        else if( m_level == 2)

            if( m_mode == MODE_FAST)
                leaf( offset, offset, offset, size, true, sections2, faces2f, face_data2f);
            else if( m_mode == MODE_EXACT)
                leaf( offset, offset, offset, size, true, sections2, faces2e, face_data2e);
            else
                assert( false);

        else
            recurse( offset, offset, offset, size, true, 0);

    if( !m_use_vbo)
        glEnd();
}

void Renderer::recurse( double x, double y, double z, double size, bool front_visible, int level)
{
    if( m_mode == MODE_STUPID && level >= m_level)

        leaf( x, y, z, size, true, sections0, faces0, face_data0);

    else if( m_mode == MODE_SIMPLE && level >= m_level) {

        if( front_visible)
            leaf( x, y, z, size, front_visible, sections0, faces0, face_data0);

    } else if( m_mode == MODE_EXACT && level >= m_level-2)

        leaf( x, y, z, size, front_visible, sections2, faces2e, face_data2e);

    else if( m_mode == MODE_FAST && level >= m_level-2)

        leaf( x, y, z, size, front_visible, sections2, faces2f, face_data2f);

    else {

       size /= 3.0;
       ++level;

       for( int i = 0; i < 20; i++) { // NOLINT modernize-loop-convert
           bool new_f = (front_visible && subcubes[i].inherited_visibility) || subcubes[i].always_visible;
           double new_x = x + subcubes[i].x*size;
           double new_y = y + subcubes[i].y*size;
           double new_z = z + subcubes[i].z*size;
           recurse( new_x, new_y, new_z, size, new_f, level);
       }
    }
}

void Renderer::leaf(
    double x, double y, double z, double size, bool front_visible, int sections, int faces, const int* face_data)
{
    double xx[sections2+1];
    double yy[sections2+1];
    double zz[sections2+1];

    size /= sections;
    for( int i=0; i <= sections; i++) {
        xx[i] = x + i*size;
        yy[i] = y + i*size;
        zz[i] = z + i*size;
    }

    // omit front faces if not visible
    if( !front_visible) {
        if( sections == 9)
            faces -= 16;
        else if( sections == 3)
            faces -= 4;
    }

    if( m_use_vbo)

        for( int i = 0; i < faces*12; i += 3) {
            m_vbo.push_back( xx[face_data[i]]);
            m_vbo.push_back( yy[face_data[i+1]]);
            m_vbo.push_back( zz[face_data[i+2]]);
        }

    else

        for( int i = 0; i < faces*12; i += 3)
            glVertex3d( xx[face_data[i]], yy[face_data[i+1]], zz[face_data[i+2]]);

    m_nr_of_quads += faces;
}

#if defined( HAVE_JPEG)
void Renderer::write_jpeg_file(
    const char* buffer, int width, int height, const std::string& filename) const
{
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;

    cinfo.err = jpeg_std_error( &jerr);
    jpeg_create_compress( static_cast<j_compress_ptr>( &cinfo));

    FILE* outfile = fopen( filename.c_str(), "wb");
    if( !outfile) {
        std::ostringstream str;
        str << "can't open " << filename;
        throw std::runtime_error( str.str());
    }

    jpeg_stdio_dest( &cinfo, outfile);

    cinfo.image_width = width;
    cinfo.image_height = height;
    cinfo.input_components = 3;
    cinfo.in_color_space = JCS_RGB;

    jpeg_set_defaults( static_cast<j_compress_ptr>( &cinfo));
    jpeg_set_quality( &cinfo, 100, TRUE);
    jpeg_start_compress( &cinfo, TRUE);

    while( cinfo.next_scanline < cinfo.image_height) {
        const char* p = &buffer[ 3 *( cinfo.image_height-cinfo.next_scanline-1) * width ];
        JSAMPROW row_pointer = reinterpret_cast<JSAMPROW>( const_cast<char*>( p));
        jpeg_write_scanlines( &cinfo, &row_pointer, 1);
    }

    jpeg_finish_compress( &cinfo);
    fclose( outfile);
    jpeg_destroy_compress( &cinfo);
}
#endif // HAVE_JPEG

#if defined( HAVE_PNG)
void Renderer::write_png_file(
    const char* buffer, int width, int height, const std::string& filename) const
{
    FILE* outfile = fopen( filename.c_str(), "wb");
    if( !outfile) {
        std::ostringstream str;
        str << "can't open " << filename;
        throw std::runtime_error( str.str());
    }

    png_structp png_ptr = png_create_write_struct( PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if( !png_ptr) {
        fclose( outfile);
        throw std::runtime_error( "can't create png_ptr");
    }

    png_infop info_ptr = png_create_info_struct( png_ptr);
    if( !info_ptr) {
        png_destroy_write_struct( static_cast<png_structpp>( &png_ptr), /*png_infopp_NULL*/ NULL);
        fclose( outfile);
        throw std::runtime_error( "can't create info_ptr");
    }

    if( setjmp( png_jmpbuf( png_ptr))) {
        png_destroy_write_struct( &png_ptr, &info_ptr);
        fclose( outfile);
        throw std::runtime_error( "png runtime error");
    }

    png_init_io( png_ptr, outfile);
    png_set_IHDR( png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_RGB,
        PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
    png_write_info( png_ptr, info_ptr);

    png_bytep row_pointers[height];
    for( int k=0; k<height; k++) {
        const char* p = &buffer[ 3 *( height-k-1) * width ];
        row_pointers[k] = reinterpret_cast<png_bytep>( const_cast<char*>( p));
    }

    png_write_image( png_ptr, row_pointers);
    png_write_end( png_ptr, info_ptr);
    png_destroy_write_struct( &png_ptr, &info_ptr);

    fclose( outfile);
}
#endif // HAVE_PNG
