//
// menger - visualization of the Menger sponge
// Copyright (C) 1995-2020  Joachim Reichel <joachim.reichel@posteo.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#ifndef MENGER_RENDERER_H
#define MENGER_RENDERER_H

#include <string>
#include <vector>

#if defined( HAVE_GLEW)
#  include <GL/glew.h>
#endif // HAVE_GLEW

#include <GL/gl.h>

/// Mode to be used for visualization.
enum Mode {

   /// Trivial recursive algorithm with 20^i quads for level i.
   MODE_STUPID,

   /// Like #MODE_SUTPID, but omits quads that are not visible because they
   /// touch an adjacent subcube. Needs 176 instead of 400 quads for level 2.
   MODE_SIMPLE,

   /// Like #MODE_SIMPLE, but uses precomputed and optimized data for level 2
   /// and below. Adjacents quads are merged to obtain larger quads, reducing
   /// the total number of quads to 96 instead of 176 for level 2.
   MODE_EXACT,

   /// Like #MODE_EXACT, but adds invisible quads to obtain ever larger and
   /// fewer quads, reducing the quads to 80 for level 2. Unfortunately, the
   /// edges of these larger quads lead to artifacts.
   MODE_FAST
};

/// This class provides a renderer for the Menger sponge.
class Renderer
{
public:

    /// Constructor.
    Renderer();

    /// Initializes the OpenGL context.
    void init();

    /// Draws the image in the current OpenGL context.
    void draw();

    /// Resizes the viewport to width \p w and height \a h.
    void resize( int w, int h) const;

#if defined( HAVE_PNG) || defined( HAVE_JPEG)
    /// Saves the image of width \p w and height \a h to file \p filename.
    void screenshot(
       int width, int height, int tile_width, int tile_height, const std::string& filename);
#endif // HAVE_PNG || HAVE_JPEG

    /// Resets the transformation.
    void reset();

    /// Translates by (\p dx, \p dy, \p dz).
    void translate( double dx, double dy, double dz);

    /// Scales by \p s around current center.
    void scale( double s);

    /// Rotates around quaternion (\p x, \p y, \p z) by \p phi.
    void rotate( double x, double y, double z, double phi);

    void set_orientation_and_position(
        double theta_x, double theta_y, double theta_z, double dx, double dy, double dz, double s);

    void get_orientation_and_position(
        double& theta_x, double& theta_y, double& theta_z, double& dx, double& dy, double& dz, double& s) const;

    /// Sets the recursion level to \p l.
    void set_level( int l);

    /// Returns the recursion level.
    int get_level() const;

    /// Sets the mode to be used to \p m.
    void set_mode( Mode m);

    /// Returns the mode to be used.
    Mode get_mode() const;

    /// Sets usage of VBOs.
    void set_use_vbo( bool v);

    /// Returns usage of VBOs.
    bool get_use_vbo() const;

    /// Returns information about current level, mode, transformation, etc.
    std::string status( int width, int height) const;

private:

#if defined( HAVE_JPEG)
    /// Saves the image as JPEG file.
    void write_jpeg_file( const char* buffer, int width, int height, const std::string& filename) const;
#endif // HAVE_JPEG

#if defined( HAVE_PNG)
    /// Saves the image as PNG file.
    void write_png_file( const char* buffer, int width, int height, const std::string& filename) const;
#endif // HAVE_PNG


    /// Generates the quad data to be rendered (one orientation only).
    void generate();

    /// Generates the quad data recusively.
    ///
    /// \param x               x coordinate of lexicographically smallest vertex
    /// \param y               y coordinate of lexicographically smallest vertex
    /// \param z               z coordinate of lexicographically smallest vertex
    /// \param size            edge length of the cube
    /// \param front_visible   indicates whether the front of the cube is
    ///                        visible
    /// \param level           recursion level of the cub
    void recurse(
        double x, double y, double z, double size, bool front_visible,
        int level);

    /// Generates the quad data (end of the recursion).
    ///
    /// \param x               x coordinate of lexicographically smallest vertex
    /// \param y               y coordinate of lexicographically smallest vertex
    /// \param z               z coordinate of lexicographically smallest vertex
    /// \param size            edge length of the cube
    /// \param front_visible   indicates whether the front of the cube is
    ///                        visible
    /// \param sections        subdivision factor for \p size used by \p face_data
    /// \param faces           length of \p face_data (divided by 12)
    /// \param face_data       12 ints per face (xyz for four quad vertices)
    void leaf(
        double x, double y, double z, double size, bool front_visible,
        int sections, int faces, const int* face_data);


    // transformation matrix
    double m_M[16];

    // recusion level
    int m_level;

    // mode to be used
    Mode m_mode;

    // number of drawn faces (per orientation)
    int m_nr_of_quads;

    // indicates whether VBOs shoud be used
    bool m_use_vbo;

    // indicates whether VBOS need to be updated
    bool m_update_vbo;

    // vertex buffer object
    std::vector<float> m_vbo;

};

#endif // MENGER_RENDERER_H
